/*
   This is a simple client that reads lines from standard input,
   sends each line as a single datagram to the filter service,
   gets back a datagram packet, reads the data from within the
   datagram packet, and then prints the datagram's data to
   standard output.

   Note: If you want to run this client without the console window
   being cluttered with so much information, then run this client
   with the following command line.

   >java UDPFilterClient  hostname  port_number  2> logFile.txt

   That will redirect stderr to a "log file" so all the extra
   information from the console window will go to the file.
*/
import java.net.*;
import java.io.*;
import java.util.Scanner;

public class UDPFilterClient
{
   private static final int DATAMAX = 255;  // Maximum size of datagram

   public static void main(String[] args) throws IOException
   {
      if (0 == args.length) // check for command line arguments
      {
         System.err.println("Usage: java UDPFilterClient <Server> [<Port>]");
         System.exit(1);
      }

      String server = args[0];  // Server name or IP address
      InetAddress serverAddress = InetAddress.getByName(server);
      int servPort = (2 == args.length) ? Integer.parseInt(args[1]) : 7;

      // Create a datagram (udp) socket to send and receive packets with.
      // This socket is bound to an ephemeral port number.
      DatagramSocket socket = new DatagramSocket();

      // Create a DatagramPacket for sending packets to the udp service.
      DatagramPacket sendPacket = new DatagramPacket(new byte[DATAMAX], DATAMAX,
                                     serverAddress, servPort);

      // Create a DatagramPacket for receiving the responses.
      DatagramPacket recvPacket = new DatagramPacket(new byte[DATAMAX], DATAMAX);

      // Create a Scanner object to make it easier to use System.in
      Scanner scanner = new Scanner( System.in );

      // Read data from stdin, put it in a packet, send packet to udp service,
      // get response packet, get data out of packet, write data to stdout.
      while ( scanner.hasNextLine() )
      {
         // Read a line from stdin.
         String oneLine = scanner.nextLine();

         // Put the line in the sending datagram packet.
         byte[] bytesToSend = oneLine.getBytes();   // allocate a new byte array
         sendPacket.setData( bytesToSend );
         sendPacket.setLength( bytesToSend.length );

         // Send the packet to the udp service.
         socket.send(sendPacket);

         // Information about the packet that was just sent.
         System.err.println( "Sent packet to " +
            sendPacket.getAddress().getHostAddress() + ":" + sendPacket.getPort() +
               " using local port " + socket.getLocalPort() );

         // Wait for a response packet (on the ephemeral port).
         socket.receive(recvPacket);

         // Information about the packet that was just received.
         System.err.println( "Received packet from " +
            recvPacket.getAddress().getHostAddress() + ":" + recvPacket.getPort() +
               " using local port " + socket.getLocalPort() );

         // Information about the data in the received packet.
         System.err.println("\trecvPacket.getLength()      = " + recvPacket.getLength() );
         System.err.println("\trecvPacket.getData().length = " + recvPacket.getData().length );

         // Get the useful data out of the received packet.
         String recvData = new String( recvPacket.getData(), 0, recvPacket.getLength() );
         System.out.println( recvData );   // write received data to stdout

         recvPacket.setLength(DATAMAX); // Reset length to avoid shrinking buffer.
                                        // (You don't want the receive() method to
                                        // only receive as many bytes as the length
                                        // of the last packet's data.)
      }

      socket.close();  // Close the socket and its streams
   }
}