/*
   This version of the client uses the same DatagramPacket object
   for both sending and receiving messages. The previous version
   used two DatagramPacket objects, one for sending and one for
   receiving (and lots of byte array objects).


   Note: If you want to run this client without the console window
   being cluttered with so much information, then run this client
   with the following command line.

   >java UDPFilterClient  hostname  port_number  2> logFile.txt

   That will redirect stderr to a "log file" so all the extra
   information from the console window will go to the file.
*/
import java.net.*;
import java.io.*;
import java.util.Scanner;

public class UDPFilterClient_v2
{
   private static final int DATAMAX = 255;  // Maximum size of datagram

   public static void main(String[] args) throws IOException
   {
      if (0 == args.length) // check for command line arguments
      {
         System.err.println("Usage: java UDPFilterClient <Server> [<Port>]");
         System.exit(1);
      }

      String server = args[0];  // Server name or IP address
      InetAddress serverAddress = InetAddress.getByName(server);
      int servPort = (2 == args.length) ? Integer.parseInt(args[1]) : 7;

      // Create a datagram (udp) socket to send and receive packets with.
      // This socket is bound to an ephemeral port number.
      DatagramSocket socket = new DatagramSocket();

      // Create a DatagramPacket for sending and receiving packets to the udp service.
      DatagramPacket packet = new DatagramPacket(new byte[DATAMAX], DATAMAX,
                                     serverAddress, servPort);

      // Create a Scanner object to make it easier to use System.in
      Scanner scanner = new Scanner( System.in );

      // Read data from stdin, put it in a packet, send packet to udp service,
      // get response packet, get data out of packet, write data to stdout.
      while ( scanner.hasNextLine() )
      {
         // Read a line from stdin.
         String oneLine = scanner.nextLine();

         // Put the line in the datagram packet.
         byte[] bytesToSend = oneLine.getBytes();   // allocate a new byte array
         packet.setData( bytesToSend );
         packet.setLength( bytesToSend.length );

         // Send the packet to the udp service.
         socket.send(packet);

         // Information about the packet that was just sent.
         System.err.println( "Sent packet to " +
            packet.getAddress().getHostAddress() + ":" + packet.getPort() +
               " using local port " + socket.getLocalPort() );

         // Wait for a response packet (on the ephemeral port).
         packet.setData( new byte[DATAMAX] );       // allocate a new byte array
         packet.setLength(DATAMAX);
         socket.receive(packet);

         // Information about the packet that was just received.
         System.err.println( "Received packet from " +
            packet.getAddress().getHostAddress() + ":" + packet.getPort() +
               " using local port " + socket.getLocalPort() );

         // Information about the data in the received packet.
         System.err.println("\trecvPacket.getLength()      = " + packet.getLength() );
         System.err.println("\trecvPacket.getData().length = " + packet.getData().length );

         // Get the useful data out of the received packet.
         String recvData = new String( packet.getData(), 0, packet.getLength() );
         System.out.println( recvData );   // write received data to stdout
      }

      socket.close();  // Close the socket and its streams
   }
}