/*
    A template for TCP clients.
    This client implements a very simple application layer protocol.
    The client receives a 1,000 character "ripple pattern" from the server
    (which is written to stdout), then sends back another ripple pattern, .
    This works well with the companion server-2-ripple.c.
*/
#include <stdio.h>      // for fprintf() and fdopen()
#include <stdlib.h>     // for atoi() and exit()
#include <string.h>     // for memset(), bzero(), bcopy(), and strerror()
#include <errno.h>      // for errno

#include <unistd.h>     // for close()
#include <sys/socket.h> // for socket(), connect(), AF_INET, SOCK_STREAM
#include <arpa/inet.h>  // for sockaddr_in, htons()
#include <netdb.h>      // for gethostbyname(), hostent, hstrerror(), and h_errno


// prototypes for two error handling functions (defined below)
void unix_error(char *msg);
void  dns_error(char *msg);


int main(int argc, char **argv)
{
   int socket_fd;                 // socket file descriptor
   struct hostent *hp;            // used to get server's IP address
   struct sockaddr_in serverAddr; // server address structure
   FILE *socketin;                // socket file pointer
   FILE *socketout;               // socket file pointer (i.e., stream)


   if (argc != 3)
   {
      fprintf(stderr, "usage: %s <host> <port>\n", argv[0]);
      exit(0);
   }

   // Create a reliable, stream socket using TCP.
   if ((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      unix_error("socket() error");

   // Construct a server address structure.
   bzero(&serverAddr, sizeof(serverAddr));        // zero out the address structure
   memset(&serverAddr, 0, sizeof(serverAddr));    // zero out the address structure
   serverAddr.sin_family = AF_INET;               // Internet address family
   if ( (hp = gethostbyname( argv[1] )) == NULL ) // fill in server's IP address
      dns_error("gethostbyname() error");
   bcopy(hp->h_addr, (struct sockaddr*)&serverAddr.sin_addr, hp->h_length);
   serverAddr.sin_port = htons( atoi(argv[2]) );  // fill in the port number


   // Establish a connection with the server.
   if (connect(socket_fd, (struct sockaddr*)&serverAddr, sizeof(serverAddr)) < 0)
      unix_error("connect() error");


   // Convert the socket file descriptor to FILE pointers
   // (i.e., streams) and use C Standard Library I/O functions.
   // (But see the textbook CS:APP, pages 796-797.)
   if ( (socketin  = fdopen(socket_fd, "r")) == NULL )
      unix_error("fdopen() error" );
   if ( (socketout = fdopen(socket_fd, "w")) == NULL )
      unix_error("fdopen() error" );

   // Notice that this client now has four open file descriptors
   // with five streams.
   // 0 STDIN_FILENO  (stdin)            half-duplex, read-only
   // 1 STDOUT_FILENO (stdout)           half-duplex, write-only
   // 2 STDERR_FILENO (stderr)           half-duplex, write-only
   // 3 socket_fd (socketin, socketout)  full-duplex, read-write

   // Implement an application level protocol.


   {  // receive a ripple pattern
      char c;
      int count = 0;
      while( 1000 > count++ )
      {
         fscanf(socketin, "%c", &c);
         printf("%c", c);
      }
   }
   {  // send a ripple pattern
      char c = 32;
      int count = 0;
      while( 1000 > count++ )
      {
         fprintf(socketout, "%c", c++);
         fflush(socketout);
         if ( c == 126 ) c = 32;
      }
   }


   // Close the socket.
   if (close(socket_fd) < 0)
      unix_error("close() error");

   exit(0);
}//main


/*
   Below are two error handling functions.
   They provide reasonably meaningful error messages.
*/
void unix_error(char *msg)
{
   fprintf(stderr, "%s: %s (errno = %d)\n", msg, strerror(errno), errno);
   exit(1);
}

void dns_error(char *msg)
{
   fprintf(stderr, "%s: %s (h_errno = %d)\n", msg, hstrerror(h_errno), h_errno);
   exit(1);
}
