/*
    A generic TCP client. This client does not implement
    any application level protocol. Instead, by using
    multiplexed I/O, this client allows a user sitting
    at the console to talk to any (ascii based) application
    layer protocol with any server.
*/
#include <stdio.h>      // for fprintf() and fdopen()
#include <stdlib.h>     // for atoi() and exit()
#include <string.h>     // for memset(), bzero(), bcopy(), and strerror()
#include <errno.h>      // for errno

#include <unistd.h>     // for close(), read(), write()
#include <sys/socket.h> // for socket(), connect(), AF_INET, SOCK_STREAM
#include <arpa/inet.h>  // for sockaddr_in, htons()
#include <netdb.h>      // for gethostbyname(), hostent, hstrerror(), and h_errno
#include <sys/select.h>

// prototypes for two error handling functions (defined below)
void unix_error(char *msg);
void  dns_error(char *msg);


int main(int argc, char **argv)
{
   int socket_fd;                 // socket file descriptor
   struct hostent *hp;            // used to get server's IP address
   struct sockaddr_in serverAddr; // server address structure

   fd_set ready_set, read_set;
   int done = 0;  // used as a boolean flag
   int countIn, countOut;
   #define BUFSIZE 32   // Size of I/O buffer
   char buf[BUFSIZE];   // I/O buffer

   if (argc != 3)
   {
      fprintf(stderr, "usage: %s <host> <port>\n", argv[0]);
      exit(0);
   }

   // Create a reliable, stream socket using TCP.
   if ((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      unix_error("socket() error");

   // Construct a server address structure.
   bzero(&serverAddr, sizeof(serverAddr));        // zero out the address structure
   memset(&serverAddr, 0, sizeof(serverAddr));    // zero out the address structure
   serverAddr.sin_family = AF_INET;               // Internet address family
   if ( (hp = gethostbyname( argv[1] )) == NULL ) // fill in server's IP address
      dns_error("gethostbyname() error");
   bcopy(hp->h_addr, (struct sockaddr*)&serverAddr.sin_addr, hp->h_length);
   serverAddr.sin_port = htons( atoi(argv[2]) );  // fill in the port number


   // Establish a connection with the server.
   if (connect(socket_fd, (struct sockaddr*)&serverAddr, sizeof(serverAddr)) < 0)
      unix_error("connect() error");

   // Set up the multiplexed I/O.
   FD_ZERO(&read_set);
   FD_SET(STDIN_FILENO, &read_set);
   FD_SET(socket_fd, &read_set);

   while(!done)
   {
      ready_set = read_set;
      if (select(socket_fd+1, &ready_set, NULL, NULL, NULL) == -1)
         unix_error("select() error");
      if (FD_ISSET(STDIN_FILENO, &ready_set))
      {
         countIn = read(STDIN_FILENO, &buf, BUFSIZE - 1);
         if (countIn < 0)
            unix_error("read() error");
         else if (countIn == 0)
            done = 1;
         else
         {
            buf[countIn-1] = '\r';  // end the buffer with CR-LF
            buf[countIn] = '\n';
            countOut = write(socket_fd, &buf, countIn+1);
            if (countOut != countIn+1)
               unix_error("write() error");
         }
      }
      if (FD_ISSET(socket_fd, &ready_set))
      {
         countIn = read(socket_fd, &buf, BUFSIZE - 1);
         if (countIn < 0)
            unix_error("read() error");
         else if (countIn == 0)
            done = 1;
         else
         {
            countOut = write(STDOUT_FILENO, &buf, countIn);
            if (countOut != countIn)
               unix_error("write() error");
         }
      }
   }

   // Close the socket.
   if (close(socket_fd) < 0)
      unix_error("close() error");

   exit(0);
}//main


/*
   Below are two error handling functions.
   They provide reasonably meaningful error messages.
*/
void unix_error(char *msg)
{
   fprintf(stderr, "%s: %s (errno = %d)\n", msg, strerror(errno), errno);
   exit(1);
}

void dns_error(char *msg)
{
   fprintf(stderr, "%s: %s (h_errno = %d)\n", msg, hstrerror(h_errno), h_errno);
   exit(1);
}
