/*
   A simple HTTP client.

   This program gets its URL from the command line.

   This version lets you set different request headers.
*/
import java.net.*;
import java.io.*;
import java.util.Date;
import java.util.TimeZone;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.nio.charset.*;

public class HTTPClient
{
   public static void main(String[] args) throws IOException
   {
      if (0 == args.length)
      {
         System.err.println("Usage: java HTTPClient <url>");
         System.exit(1);
      }

      URL url = null;
      String arg = args[0].trim();
      try // Create a URL object using the command line argument.
      {
         url = new URL( (arg.startsWith("http://")?"":"http://") + arg );
      }
      catch (MalformedURLException e)
      {
         System.err.println("Malformed URL: " + (arg.startsWith("http://")?"":"http://") + arg );
         e.printStackTrace(System.err);
         System.exit(1);
      }

      // Set the port number.
      int servPort = url.getPort();
      if ( -1 == servPort )
         servPort = 80;

      // Create a Socket object that is connected to the server on the specified port.
      Socket socket =  null;
      try
      {
         socket = new Socket(url.getHost(), servPort);
      }
      catch (UnknownHostException e)
      {
         System.err.println("Bad URL: " + url );
         e.printStackTrace(System.err);
         System.exit(1);
      }
      // Log information about the connection that was just established.
      System.err.println( "Connected to server " +
         socket.getInetAddress().getHostAddress() + ":" + socket.getPort() +
            " using local port " + socket.getLocalPort() + "\n" );

      // Get (low-level) input and output streams from the socket.
      InputStream   in = socket.getInputStream();
      OutputStream out = socket.getOutputStream();

      // Make the low-level input and output streams easier to use.
      PrintStream ps = new PrintStream( out );
      // NOTE: DataInputStream is not really the best stream to use
      // (its readLine() method is deprecated) but it makes it easy
      // to read both line oriented data (ie, headers) and binary data.
      DataInputStream br = new DataInputStream(new BufferedInputStream( in ));
      // NOTE: BufferedReader is the right stream to use for line
      // oriented data (like headers) but it will mangle any binary data
      // (like compressed files or image files). Try it and see what happens.
    //BufferedReader br = new BufferedReader( new InputStreamReader(in, "latin1") );


      // Use only one of the following two request methods.
      String requestMethod1 = "GET";
      String requestMethod2 = "HEAD";
      // Use only one of the following http versions.
      String httpVersion1 = "HTTP/1.1";
      String httpVersion0 = "HTTP/1.0";  // NOTE: Use HTTP/1.0 to prevent chunking.
      String httpVersion9 = "HTTP/0.9";

      // Choose the request method and http version
      String requestMethod = requestMethod1;
      String httpVersion   = httpVersion1;

      // Build the request line.
      String request = "";
      request += requestMethod + " ";
      request += ( url.getPath().isEmpty()?"/":url.getPath() ) + " ";
      request += httpVersion   + "\r\n";


      // Create three dates that we can use in the request headers.
      //  www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.3.1
      //  stackoverflow.com/questions/7707555/getting-date-in-http-format-in-java
      // First, create an object that formats dates into "HTTP-date" format.
      DateFormat df = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss z");
      df.setTimeZone( TimeZone.getTimeZone("GMT") ); // the required time zone
      // Format three dates for use in the request headers.
      String    now = df.format( new Date() );
      String future = df.format( new Date((new Date()).getTime() + 11000) );
      String   past = df.format( new Date(0)); // January 1, 1970, 00:00:00 GMT


      // Build the request headers.
      request += "host: " + url.getHost() + (80==servPort?"":":"+servPort) + "\r\n";
      request += "User-Agent: HTTPClient.java\r\n";
      request += "Date: " + now + "\n";
      request += "Accept-Language: en-US,en;q=0.8\r\n";
      request += "Accept: text/html,text/css,image/webp,*/*;q=0.8\r\n";
      // Use one of these two headers to turn on compressed entity bodies.
      request += "Accept-Encoding: gzip\r\n";
    //request += "Accept-Encoding: deflate\r\n";
      // Use only one (or neither) of the following three headers.
    //request += "If-Modified-Since: " + past + "\r\n";
      request += "If-Modified-Since: " + now  + "\r\n";
    //request += "If-Modified-Since: " + future  + "\r\n";
      // Use only one (or neither) of the following two headers.
    //request += "Connection: close\r\n";
      request += "Connection: keep-alive\r\n";  // default for HTTP/1.1
      request += "\r\n";  // blank line

      // Send the request to the server.
      ps.print(request);
      ps.flush();
      // Log the request.
      System.out.println(request);


      // Receive the response line from the server.
      String responseLine = br.readLine();
      // Log the response line.
      System.out.println( responseLine );

      // Receive (and parse) the response headers from the server.
      String responseHeader = null;
      String contentTypeExtension = null;
      boolean contentTypeText = true;  // default to text
      boolean contentGzipped  = false; // default to not gzipped
      boolean contentZzipped  = false; // default to not zlibed
      boolean transferChunked = false; // default to not chunked
      int contentLength = 1000000;     // default value for entity buffer size
      while (null != (responseHeader = br.readLine()))
      {
         if ( responseHeader.startsWith("Content-Length:") )
         {
            int i = responseHeader.indexOf(':');
            contentLength = Integer.parseInt(responseHeader.substring(i+2));
         }
         else if ( responseHeader.startsWith("Content-Type:") )
         {
            int i = responseHeader.indexOf(':');
            contentTypeText = responseHeader.substring(i+1).trim().startsWith("text");
            int i2 = responseHeader.indexOf('/');
            int i3 = responseHeader.indexOf(';');
            if (-1 != i3)
               contentTypeExtension = responseHeader.substring(1+i2, i3);
            else
               contentTypeExtension = responseHeader.substring(1+i2);
         }
         else if ( responseHeader.startsWith("Content-Encoding:") )
         {
            int i = responseHeader.indexOf(':');
            contentGzipped = responseHeader.substring(i+1).trim().startsWith("gzip");
            contentZzipped = responseHeader.substring(i+1).trim().startsWith("deflate");
         }
         else if ( responseHeader.startsWith("Transfer-Encoding: chunked") )
         {
            transferChunked = true;
         }
         // Log the response header.
         System.out.println(responseHeader);
         if ( responseHeader.isEmpty() ) // a blank line
            break;
      }

      // Receive the entity body from the server.
    //char[] entityBuffer = new char[contentLength];  // use this with BufferedReader
      byte[] entityBuffer = new byte[contentLength];  // use this with DataInputStream
      int totalBytesRcvd = 0;  // Total bytes received so far
      int bytesRcvd;           // Bytes received in last read
      while ( (totalBytesRcvd < contentLength)
           && (-1 != (bytesRcvd = br.read(entityBuffer,
                                          totalBytesRcvd,
                                          contentLength - totalBytesRcvd))))
      {
         totalBytesRcvd += bytesRcvd;
      }

      // Send the response body to standard output.
      if ( requestMethod.equals("GET")
        && contentTypeText
        && !contentGzipped
        && !contentZzipped
        && 0 < totalBytesRcvd && totalBytesRcvd < 10000 )
      {
         System.out.println( new String(entityBuffer, 0, totalBytesRcvd) );
      }
      else if (requestMethod.equals("GET"))
      {
         System.out.println("Received " + totalBytesRcvd + " bytes of data.");
      }

      // Save the response body as a file (so it can be inspected).
      if (contentTypeExtension != null)
      {
         String fileName = "temp" + "." + (transferChunked?"chunked.":"") + contentTypeExtension;
         if ( contentGzipped ) fileName += ".gz";
         if ( contentZzipped ) fileName += ".zlib";
         // Use these three lines with DataInputStream.
         FileOutputStream fos = new FileOutputStream(fileName);
         fos.write(entityBuffer, 0, totalBytesRcvd);
         fos.close();
         // Use these three lines with BufferedReader (the character stream).
       //FileWriter fw = new FileWriter(fileName);
       //fw.write(entityBuffer, 0, totalBytesRcvd);
       //fw.close();
         System.out.println("Entity body saved as file: " + fileName );
      }

      br.close();  // Close the socket and its streams
      ps.close();
      socket.close();
      System.out.println( "connection closed." );
   }
}