import java.io.*;
import java.net.*;
import java.util.*;

class WebServer
{
   private static String serverRoot = "html"; // This should be a configuration option.
   private static int portNumber;             // This is set by a command line argument.

   public static void main(String[] args) throws Exception
   {
      // Check for a command line argument.
      if (0 == args.length)
      {
         System.err.println("Usage: java WebServer <port>");
         System.exit(1);
      }
      portNumber = Integer.parseInt(args[0]);

      // Create a server socket to accept client connection requests.
      ServerSocket listenSocket = new ServerSocket( portNumber );

      System.err.println("****** WebSever waiting on port " +
                                 portNumber + " for client connections.");

      while (true) // Run forever, accepting and servicing clients.
      {
         Socket clntSock = listenSocket.accept();

         // Log information about the connection that was just established.
         System.err.println("****** Handling client " +
            clntSock.getInetAddress().getCanonicalHostName() + " at " +
               clntSock.getInetAddress().getHostAddress() + ":" + clntSock.getPort());

         handleConnection(clntSock);
      }
   }//main()



   public static void handleConnection(Socket clntSock) throws Exception
   {
      // Get (low-level) input and output streams from the socket.
      InputStream   in = clntSock.getInputStream();
      OutputStream out = clntSock.getOutputStream();

      // Make the low-level input and output streams easier to use.
      BufferedReader inFromClient = new BufferedReader(new InputStreamReader(in));
      PrintStream     outToClient = new PrintStream( out );

      // Read the request line.
      String requestLine = inFromClient.readLine();
      // Sometimes, I don't know why, the result line is null.
      if (null == requestLine)
      {
         System.err.println( "=====> null Request Line\n" );
         clntSock.shutdownInput();
         inFromClient.close();
         outToClient.close();
         clntSock.close();
         return;  // Finished with this client.
      }
      // Log the request line.
      System.err.println( requestLine );

      //**** Parse the request line.

      requestLine = requestLine.trim();
      // Check for a proper version of HTTP in the request line.
      String temp1 = requestLine.toUpperCase();
      if ( ! ( temp1.endsWith(" HTTP/1.1")
            || temp1.endsWith(" HTTP/1.0")
            || temp1.endsWith(" HTTP/0.9") ) )
      {
         sendErrorResponse(outToClient,
             "505", "HTTP Version Not Supported",
             "The HTTP version in this request<br>" + requestLine +
                  "<br>is unsupported.");
         clntSock.shutdownInput();
         inFromClient.close();
         outToClient.close();
         clntSock.close();
         return;  // Finished with this client.
      }
      // Remove the HTTP version from the request line.
      int index1 = requestLine.lastIndexOf(" ");
      String httpVersion = requestLine.substring(1+index1).toUpperCase();
      requestLine = requestLine.substring(0, index1).trim();

      // Check that the request line begins with a supported method.
      String temp2 = requestLine.toUpperCase();
      if ( ! ( temp2.startsWith("GET ")
            || temp2.startsWith("POST ") ) )
      {
         sendErrorResponse(outToClient, "501", "Not Implemented",
             "The HTTP method in this request<br>" + requestLine +
                  "<br>is not supported.");
         clntSock.shutdownInput();
         inFromClient.close();
         outToClient.close();
         clntSock.close();
         return;  // Finished with this client.
      }
      // Break the request line into a request method and a URL.
      int index2 = requestLine.indexOf(" ");
      String requestMethod = requestLine.substring(0, index2).toUpperCase();
      String requestURL = requestLine.substring(1+index2).trim();

      // Important: We need to decode the URL from its urlencoded form.
      // See
      //   http://www.w3schools.com/tags/ref_urlencode.asp
      // or
      //   http://en.wikipedia.org/wiki/URL_encoding
      String decodedURL = URLDecoder.decode(requestURL, "UTF-8");
/*
      // Log the results of decoding the URL.
      System.err.println( "=====> requestURL = " + requestURL );
      System.out.println( "=====> decodedURL = " + decodedURL );
*/
      // Break the decoded URL into a resource and a query.
      String resourceName = null;
      String queryString = null;
      if ( 0 < decodedURL.indexOf('?') )
      {
         String[] temp = decodedURL.split("\\Q?\\E");
         resourceName  = temp[0];
         queryString   = temp[1];
      }
      else
      {
         resourceName = decodedURL;
      }

      // Log the results of parsing the request line.
      System.err.println( "=====> Request method = " + requestMethod );
      System.out.println( "=====> Resource name = " + resourceName );
      System.err.println( "=====> Query string = " + queryString );
      System.err.println( "=====> HTTP version = " + httpVersion );

      //**** Done parsing the request line.

      // Important: Check that the resource is not a file
      // that "escapes" from the server's root directory.
      // This is a form of server attack that tries to
      // sneek in a URL like this.
      //           /../../../../Windows/System32/cmd.exe
      File file1 = new File(serverRoot).getCanonicalFile();
      File file2 = new File(serverRoot + resourceName).getCanonicalFile();
      if (! file2.getAbsolutePath().startsWith(file1.getAbsolutePath()))
      {
         sendErrorResponse(outToClient, "403", "Forbidden",
             "You can't pull this stunt<br>"+requestLine+"<br>on us.");
         clntSock.shutdownInput();
         inFromClient.close();
         outToClient.close();
         clntSock.close();
         return;  // Finished with this client.
      }

      // Make sure the resource exits and is readable.
      File file = new File( serverRoot + resourceName );
      if ( ! file.exists() )
      {
         sendErrorResponse(outToClient, "404", "Not Found",
             "The requested URL "+resourceName+" was not found on this server.");
         clntSock.shutdownInput();
         inFromClient.close();
         outToClient.close();
         clntSock.close();
         return;  // Finished with this client.
      }
      else if ( ! file.canRead() )
      {
         sendErrorResponse(outToClient, "403", "Forbidden",
             "The requested URL "+resourceName+" is not readable.");
         clntSock.shutdownInput();
         inFromClient.close();
         outToClient.close();
         clntSock.close();
         return;  // Finished with this client.
      }

      // Process the request method.
      if ( requestMethod.equals( "GET" ) )
      {
         if ( file.isDirectory() )
         {
            doDynDirListing(outToClient, resourceName, file);
         }
         else if ( resourceName.endsWith(".exe") )
         {
            doCGIGet(outToClient, queryString, resourceName, file);
         }
         else // the file exists, so send it to the client
         {
            doStaticGet(inFromClient, outToClient, resourceName, file);
         }
      }
      else if ( requestMethod.equals( "POST" ) )
      {
         if ( file.isDirectory() )
         {
            sendErrorResponse(outToClient, "405", "Method Not Allowed",
                "The POST method applied to a directory<br>" + resourceName +
                      "<br>is not allowed.");
         }
         else if ( resourceName.endsWith(".exe") )
         {
            doCGIPost(inFromClient, outToClient, resourceName, file);
         }
         else
         {
            doStaticPost(inFromClient, outToClient, resourceName, file);
         }
      }
      else  // unsupported request method
      {
         sendErrorResponse(outToClient, "501", "Not Implemented",
             "This server does not implement this method: " + requestMethod);
      }

      // Close the streams and the communicating socket.
      inFromClient.close();
      outToClient.close();
      clntSock.close();
      return;
   }//handleConnection()



   public static void doStaticGet(BufferedReader inFromClient,
                                  PrintStream outToClient,
                                  String fileName,
                                  File file)
   throws Exception
   {
      // Read all of the request headers. (Note: We do not make
      // any use of these request headers. A more sophisticated
      // implementation of GET would need to make use of them.)
      String requestHeader;
      while ( (requestHeader = inFromClient.readLine()) != null )
      {
         // Log the request header.
         System.err.println( requestHeader );
         if ( requestHeader.isEmpty() )  // stop on a blank line
         {
            break;
         }
      }

      int numOfBytes = (int) file.length();
      FileInputStream inFile = new FileInputStream( file );
      byte[] fileInBytes = new byte[numOfBytes];
      inFile.read(fileInBytes);

      // Build the HTTP response line and headers.
      // The response line.
      String response = "HTTP/1.1 200 OK\r\n";
      // The response headers.
      response += "Connection: close\r\n";
      if ( fileName.endsWith(".html") || fileName.endsWith(".htm") )
         response += "Content-Type: text/html\r\n";
      else if ( fileName.endsWith(".css") )
         response += "Content-Type: text/css\r\n";
      else if ( fileName.endsWith(".js") )
         response += "Content-Type: application/javascript\r\n";
      else if ( fileName.endsWith(".jpg") || fileName.endsWith(".jpeg") )
         response += "Content-Type: image/jpeg\r\n";
      else if ( fileName.endsWith(".gif") )
         response += "Content-Type: image/gif\r\n";
      else if ( fileName.endsWith(".png") )
         response += "Content-Type: image/png\r\n";
      else  // default content type
         response += "Content-Type: text/plain\r\n";
      response += "Content-Length: " + numOfBytes + "\r\n";
      // Add a blank line to denote end of response headers.
      response += "\r\n";

      // Log the response.
      System.err.print( response );
      // Send the response.
      outToClient.print( response );
      // Send the entity body.
      outToClient.write(fileInBytes, 0, numOfBytes);
   }//doStaticGet()



   public static void doStaticPost(BufferedReader inFromClient,
                                   PrintStream outToClient,
                                   String fileName,
                                   File file)
   throws Exception
   {
      // This is just a sketch of an implementation of the POST method.
      // This sketch reads the request headers looking for the size of the
      // entity body, then it reads the entity body, prints it to standard
      // error, and then sends back to the client a "thank you" page.
      // This implementation of POST ignores the resource given in the URL.

      // Read the request headers, look for "Content-Length".
      int entityLength = 0;
      String requestHeader;
      while ( (requestHeader = inFromClient.readLine()) != null )
      {
         // Log the request header.
         System.err.println( requestHeader );
         if ( requestHeader.startsWith("Content-Length") )
         {
            int i = requestHeader.indexOf(':');
            entityLength = Integer.parseInt(requestHeader.substring(i+2));
            // Log the value of Content-Length
            //System.err.println("=====> Content-Length = " + entityLength);
         }
         if ( requestHeader.isEmpty() )  // stop on a blank line
         {
            break;
         }
      }

      // Read the entity body.
      char[] entityBodyChars = new char[entityLength];
      int totalBytesRcvd = 0;  // Total bytes received so far.
      int bytesRcvd;           // Bytes received in last read.
      while ( totalBytesRcvd < entityLength )
      {
         System.err.println("=====> Reading entity body.");
         bytesRcvd = inFromClient.read(entityBodyChars,
                                       totalBytesRcvd,
                                       entityLength - totalBytesRcvd);
         if (-1 == bytesRcvd)
         {
            break; // reached end-of-file
         }
         totalBytesRcvd += bytesRcvd;
         System.err.println("=====> totalBytesRcvd = " + totalBytesRcvd);
      }

      // Log the entity body to stderr.
      System.err.println( "=====> " + new String(entityBodyChars) + "\n" );

      // Build the HTTP response body.
      String body = "<!doctype html>\r\n" +
                    "<html>\r\n" +
                    "<head>\r\n" +
                    "<title>Thank you!</title>\r\n" +
                    "</head>\r\n" +
                    "<body>\r\n" +
                    "<h1>Thank you!</h1>\r\n" +
                    "<p>Thank you for your data!</p>\r\n" +
                    "<p>We will protect it forever.</p>\r\n" +
                    "</body>\r\n" +
                    "</html>\r\n";

      // Build the HTTP response line and headers.
      // The response line.
      String response = "HTTP/1.1 200 OK\r\n";
      // The response headers.
      response += "Connection: close\r\n";
      response += "Content-type: text/html\r\n";
      response += "Content-length: " + body.length() + "\r\n";
      // Add a blank line to denote end of response headers.
      response += "\r\n";

      // Log the response.
      System.err.print( response );
      // Send the response.
      outToClient.print( response );
      // Send the entity body.
      outToClient.print( body );
   }//doStaticPost()



   public static void sendErrorResponse(PrintStream outToClient,
                                        String code,
                                        String shortMsg,
                                        String longMsg)
   throws Exception
   {
      // Build the response body.
      String body = "<!doctype html>\r\n" +
                    "<html>\r\n" +
                    "<head>\r\n" +
                    "<title>" + code + " " + shortMsg + "</title>\r\n" +
                    "</head>\r\n" +
                    "<body>\r\n" +
                    "<h1>" + shortMsg + "</h1>\r\n" +
                    "<p>" + code + ": " + shortMsg + "</p>\r\n" +
                    "<p>" + longMsg + "</p>\r\n" +
                    "<hr>\r\n" +
                    "<address>WebServer.java at " +
                        InetAddress.getLocalHost().getCanonicalHostName() +
                        " Port " + portNumber +
                    "</address>\r\n" +
                    "</body>\r\n" +
                    "</html>\r\n";

      // Build the HTTP response line and headers.
      // The response line.
      String response = "HTTP/1.1 " + code + " " + shortMsg + "\r\n";
      // The response headers.
      response += "Connection: close\r\n";
      response += "Content-type: text/html\r\n";
      response += "Content-length: " + body.length() + "\r\n";
      // Add a blank line to denote end of response headers.
      response += "\r\n";

     // Log the response.
      System.err.print( response );
      // Send the response.
      outToClient.print( response );
      // Send the entity body.
      outToClient.print( body );
   }//sendErrorResponse()



   /**** This method should implement the CGI POST protocol for
         inter-process communication between the server process,
         its child process, and the client process.
   */
   public static void doCGIPost(BufferedReader inFromClient,
                                PrintStream outToClient,
                                String fileName,
                                File file)
   throws Exception
   {
      sendErrorResponse(outToClient, "400", "Bad Request",
          "This server cannot yet handle CGI POST requests.");
   }



   /**** This method should implement the CGI GET protocol for
         inter-process communication between the server process,
         its child process, and the client process.
   */
   public static void doCGIGet(PrintStream outToClient,
                               String queryString,
                               String fileName,
                               File file)
   throws Exception
   {
      sendErrorResponse(outToClient, "400", "Bad Request",
          "This server cannot yet handle CGI GET requests.");
   }



   /**** This method should be implemented to send the client a
         html page that is a listing of the given directory.
   */
   public static void doDynDirListing(PrintStream outToClient,
                                      String fileName,
                                      File file)
   throws Exception
   {
      sendErrorResponse(outToClient, "400", "Bad Request",
          "The requested URL, "+fileName+", does not contain a file name.");
   }
}